"""Tests for copyContents module."""

import os
import shutil
import tempfile
from unittest import mock

import pytest

from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.tasks.utility.copyContents import (
    ConsolidateUnpackagedMetadata,
    clean_temp_directory,
    consolidate_metadata,
    copy_directory_contents,
    copy_item_to_destination,
    copy_matched_files,
    merge_directory_contents,
    print_directory_tree,
    resolve_file_pattern,
    validate_directory,
)


class TestMergeDirectoryContents:
    """Test cases for merge_directory_contents function."""

    def test_merge_empty_directories(self):
        """Test merging two empty directories."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            merge_directory_contents(src_dir, dest_dir)
            assert os.path.exists(dest_dir)

    def test_merge_files_from_source(self):
        """Test merging files from source to destination."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create file in source
            src_file = os.path.join(src_dir, "test.txt")
            with open(src_file, "w") as f:
                f.write("source content")

            merge_directory_contents(src_dir, dest_dir)

            dest_file = os.path.join(dest_dir, "test.txt")
            assert os.path.exists(dest_file)
            with open(dest_file, "r") as f:
                assert f.read() == "source content"

    def test_merge_overwrites_existing_files(self):
        """Test that source files overwrite destination files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create file in destination
            dest_file = os.path.join(dest_dir, "test.txt")
            with open(dest_file, "w") as f:
                f.write("old content")

            # Create file in source with different content
            src_file = os.path.join(src_dir, "test.txt")
            with open(src_file, "w") as f:
                f.write("new content")

            merge_directory_contents(src_dir, dest_dir)

            with open(dest_file, "r") as f:
                assert f.read() == "new content"

    def test_merge_nested_directories(self):
        """Test merging nested directory structures."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create nested structure in source
            nested_src = os.path.join(src_dir, "nested", "subdir")
            os.makedirs(nested_src)
            nested_file = os.path.join(nested_src, "file.txt")
            with open(nested_file, "w") as f:
                f.write("nested content")

            merge_directory_contents(src_dir, dest_dir)

            nested_dest = os.path.join(dest_dir, "nested", "subdir", "file.txt")
            assert os.path.exists(nested_dest)
            with open(nested_dest, "r") as f:
                assert f.read() == "nested content"

    def test_merge_with_overwrite_flag(self):
        """Test merge with overwrite flag enabled."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create file in destination
            dest_file = os.path.join(dest_dir, "test.txt")
            with open(dest_file, "w") as f:
                f.write("old")

            # Create directory with same name in source
            src_subdir = os.path.join(src_dir, "test.txt")
            os.makedirs(src_subdir)

            merge_directory_contents(src_dir, dest_dir, overwrite=True)

            # Should be replaced with directory
            assert os.path.isdir(os.path.join(dest_dir, "test.txt"))


class TestCopyItemToDestination:
    """Test cases for copy_item_to_destination function."""

    def test_copy_file_to_new_location(self):
        """Test copying a file to a new location."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_file = os.path.join(tmpdir, "source.txt")
            dest_file = os.path.join(tmpdir, "dest.txt")

            with open(src_file, "w") as f:
                f.write("test content")

            copy_item_to_destination(src_file, dest_file)

            assert os.path.exists(dest_file)
            with open(dest_file, "r") as f:
                assert f.read() == "test content"

    def test_copy_directory_to_new_location(self):
        """Test copying a directory to a new location."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)

            src_file = os.path.join(src_dir, "file.txt")
            with open(src_file, "w") as f:
                f.write("content")

            copy_item_to_destination(src_dir, dest_dir)

            assert os.path.exists(dest_dir)
            assert os.path.exists(os.path.join(dest_dir, "file.txt"))

    def test_copy_directory_merges_with_existing(self):
        """Test copying directory merges with existing directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # File in source
            src_file = os.path.join(src_dir, "src_file.txt")
            with open(src_file, "w") as f:
                f.write("src")

            # File in destination
            dest_file = os.path.join(dest_dir, "dest_file.txt")
            with open(dest_file, "w") as f:
                f.write("dest")

            copy_item_to_destination(src_dir, dest_dir)

            # Both files should exist
            assert os.path.exists(os.path.join(dest_dir, "src_file.txt"))
            assert os.path.exists(os.path.join(dest_dir, "dest_file.txt"))

    def test_copy_file_overwrites_existing_file(self):
        """Test copying file overwrites existing file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_file = os.path.join(tmpdir, "source.txt")
            dest_file = os.path.join(tmpdir, "dest.txt")

            with open(dest_file, "w") as f:
                f.write("old")

            with open(src_file, "w") as f:
                f.write("new")

            copy_item_to_destination(src_file, dest_file)

            with open(dest_file, "r") as f:
                assert f.read() == "new"

    def test_copy_file_replaces_directory_with_overwrite(self):
        """Test copying file replaces directory when overwrite=True."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_file = os.path.join(tmpdir, "source.txt")
            dest_dir = os.path.join(tmpdir, "dest")

            os.makedirs(dest_dir)
            with open(src_file, "w") as f:
                f.write("file content")

            copy_item_to_destination(src_file, dest_dir, overwrite=True)

            assert os.path.isfile(dest_dir)
            with open(dest_dir, "r") as f:
                assert f.read() == "file content"

    def test_copy_directory_replaces_file_with_overwrite(self):
        """Test copying directory replaces file when overwrite=True."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_file = os.path.join(tmpdir, "dest")

            os.makedirs(src_dir)
            src_file = os.path.join(src_dir, "file.txt")
            with open(src_file, "w") as f:
                f.write("content")

            with open(dest_file, "w") as f:
                f.write("old")

            copy_item_to_destination(src_dir, dest_file, overwrite=True)

            assert os.path.isdir(dest_file)
            assert os.path.exists(os.path.join(dest_file, "file.txt"))


class TestCopyDirectoryContents:
    """Test cases for copy_directory_contents function."""

    def test_copy_all_contents(self):
        """Test copying all contents from source to destination."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create files and subdirectories
            with open(os.path.join(src_dir, "file1.txt"), "w") as f:
                f.write("file1")
            os.makedirs(os.path.join(src_dir, "subdir"))
            with open(os.path.join(src_dir, "subdir", "file2.txt"), "w") as f:
                f.write("file2")

            copy_directory_contents(src_dir, dest_dir)

            assert os.path.exists(os.path.join(dest_dir, "file1.txt"))
            assert os.path.exists(os.path.join(dest_dir, "subdir", "file2.txt"))

    def test_extract_src_directory(self):
        """Test extracting src directory contents."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create src subdirectory
            src_subdir = os.path.join(src_dir, "src")
            os.makedirs(src_subdir)
            with open(os.path.join(src_subdir, "file.txt"), "w") as f:
                f.write("content")

            copy_directory_contents(src_dir, dest_dir, extract_src=True)

            # File should be in dest_dir/src/file.txt
            assert os.path.exists(os.path.join(dest_dir, "src", "file.txt"))

    def test_extract_src_with_no_src_directory(self):
        """Test extract_src when src directory doesn't exist."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            with open(os.path.join(src_dir, "file.txt"), "w") as f:
                f.write("content")

            copy_directory_contents(src_dir, dest_dir, extract_src=True)

            # Should copy normally since no src directory
            assert os.path.exists(os.path.join(dest_dir, "file.txt"))

    def test_copy_with_overwrite(self):
        """Test copying with overwrite flag."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create conflicting file
            dest_file = os.path.join(dest_dir, "test.txt")
            with open(dest_file, "w") as f:
                f.write("old")

            src_file = os.path.join(src_dir, "test.txt")
            with open(src_file, "w") as f:
                f.write("new")

            copy_directory_contents(src_dir, dest_dir, overwrite=True)

            with open(dest_file, "r") as f:
                assert f.read() == "new"


class TestResolveFilePattern:
    """Test cases for resolve_file_pattern function."""

    def test_resolve_specific_file(self):
        """Test resolving a specific file path."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_file = os.path.join(tmpdir, "test.txt")
            with open(test_file, "w") as f:
                f.write("content")

            result = resolve_file_pattern("test.txt", tmpdir)
            assert len(result) == 1
            assert result[0] == test_file

    def test_resolve_glob_pattern(self):
        """Test resolving a glob pattern."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create multiple files
            for i in range(3):
                with open(os.path.join(tmpdir, f"file{i}.txt"), "w") as f:
                    f.write(f"content{i}")

            result = resolve_file_pattern("*.txt", tmpdir)
            assert len(result) == 3

    def test_resolve_recursive_pattern(self):
        """Test resolving recursive glob pattern."""
        with tempfile.TemporaryDirectory() as tmpdir:
            subdir = os.path.join(tmpdir, "subdir")
            os.makedirs(subdir)

            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(subdir, "file2.txt"), "w") as f:
                f.write("content2")

            result = resolve_file_pattern("**/*.txt", tmpdir)
            assert len(result) == 2

    def test_resolve_pattern_no_match_raises_error(self):
        """Test that resolving non-existent pattern raises ValueError."""
        with tempfile.TemporaryDirectory() as tmpdir:
            result = resolve_file_pattern("nonexistent.txt", tmpdir)
            assert len(result) == 0
            assert result == []

    def test_resolve_pattern_with_subdirectory(self):
        """Test resolving pattern in subdirectory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            subdir = os.path.join(tmpdir, "subdir")
            os.makedirs(subdir)
            test_file = os.path.join(subdir, "test.txt")
            with open(test_file, "w") as f:
                f.write("content")

            result = resolve_file_pattern("subdir/test.txt", tmpdir)
            assert len(result) == 1
            assert result[0] == test_file


class TestCopyMatchedFiles:
    """Test cases for copy_matched_files function."""

    def test_copy_single_file(self):
        """Test copying a single matched file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            source_file = os.path.join(source_dir, "file.txt")
            with open(source_file, "w") as f:
                f.write("content")

            copy_matched_files([source_file], source_dir, dest_dir)

            dest_file = os.path.join(dest_dir, "file.txt")
            assert os.path.exists(dest_file)
            with open(dest_file, "r") as f:
                assert f.read() == "content"

    def test_copy_preserves_relative_structure(self):
        """Test copying preserves relative directory structure."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            nested_file = os.path.join(source_dir, "subdir", "file.txt")
            os.makedirs(os.path.dirname(nested_file))
            with open(nested_file, "w") as f:
                f.write("content")

            copy_matched_files([nested_file], source_dir, dest_dir)

            dest_file = os.path.join(dest_dir, "subdir", "file.txt")
            assert os.path.exists(dest_file)

    def test_copy_multiple_files(self):
        """Test copying multiple matched files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            files = []
            for i in range(3):
                file_path = os.path.join(source_dir, f"file{i}.txt")
                with open(file_path, "w") as f:
                    f.write(f"content{i}")
                files.append(file_path)

            copy_matched_files(files, source_dir, dest_dir)

            for i in range(3):
                assert os.path.exists(os.path.join(dest_dir, f"file{i}.txt"))


class TestCleanTempDirectory:
    """Test cases for clean_temp_directory function."""

    def test_clean_existing_directory(self):
        """Test cleaning an existing temporary directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test")
            os.makedirs(test_dir)

            with open(os.path.join(test_dir, "file.txt"), "w") as f:
                f.write("content")

            clean_temp_directory(test_dir)
            assert not os.path.exists(test_dir)

    def test_clean_nonexistent_directory(self):
        """Test cleaning a non-existent directory doesn't raise error."""
        clean_temp_directory("/nonexistent/path")
        # Should not raise an exception

    def test_clean_directory_with_nested_structure(self):
        """Test cleaning directory with nested structure."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test")
            nested = os.path.join(test_dir, "nested", "deep")
            os.makedirs(nested)

            with open(os.path.join(nested, "file.txt"), "w") as f:
                f.write("content")

            clean_temp_directory(test_dir)
            assert not os.path.exists(test_dir)


class TestValidateDirectory:
    """Test cases for validate_directory function."""

    def test_validate_existing_directory(self):
        """Test validating an existing directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            validate_directory(tmpdir)
            # Should not raise

    def test_validate_nonexistent_path_raises_error(self):
        """Test validating non-existent path raises ValueError."""
        with pytest.raises(ValueError, match="does not exist"):
            validate_directory("/nonexistent/path")

    def test_validate_file_raises_error(self):
        """Test validating a file raises ValueError."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_file = os.path.join(tmpdir, "test.txt")
            with open(test_file, "w") as f:
                f.write("content")

            with pytest.raises(ValueError, match="is not a directory"):
                validate_directory(test_file)

    def test_validate_with_custom_path_name(self):
        """Test validating with custom path name in error message."""
        with pytest.raises(ValueError, match="custom_name does not exist"):
            validate_directory("/nonexistent", "custom_name")


class TestConsolidateMetadata:
    """Test cases for consolidate_metadata function."""

    def test_consolidate_string_path(self):
        """Test consolidating metadata from string path."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content")

            result, file_count = consolidate_metadata("source", tmpdir)

            assert os.path.exists(result)
            assert os.path.exists(os.path.join(result, "file.txt"))
            assert file_count == 1
            clean_temp_directory(result)

    def test_consolidate_list_paths(self):
        """Test consolidating metadata from list of paths."""
        with tempfile.TemporaryDirectory() as tmpdir:
            dir1 = os.path.join(tmpdir, "dir1")
            dir2 = os.path.join(tmpdir, "dir2")
            os.makedirs(dir1)
            os.makedirs(dir2)

            with open(os.path.join(dir1, "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(dir2, "file2.txt"), "w") as f:
                f.write("content2")

            result, file_count = consolidate_metadata(["dir1", "dir2"], tmpdir)

            assert os.path.exists(os.path.join(result, "file1.txt"))
            assert os.path.exists(os.path.join(result, "file2.txt"))
            assert file_count == 2
            clean_temp_directory(result)

    def test_consolidate_dict_with_wildcard(self):
        """Test consolidating metadata from dict with wildcard pattern."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            src_subdir = os.path.join(source_dir, "src")
            os.makedirs(src_subdir)

            with open(os.path.join(src_subdir, "file.txt"), "w") as f:
                f.write("content")

            result, file_count = consolidate_metadata({"source": "*.*"}, tmpdir)

            assert os.path.exists(os.path.join(result, "src", "file.txt"))
            assert file_count == 1
            clean_temp_directory(result)

    def test_consolidate_dict_with_single_pattern(self):
        """Test consolidating metadata from dict with single pattern."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            test_file = os.path.join(source_dir, "test.txt")
            with open(test_file, "w") as f:
                f.write("content")

            result, file_count = consolidate_metadata({"source": "test.txt"}, tmpdir)

            assert os.path.exists(os.path.join(result, "test.txt"))
            assert file_count == 1
            clean_temp_directory(result)

    def test_consolidate_dict_with_list_patterns(self):
        """Test consolidating metadata from dict with list of patterns."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            file1 = os.path.join(source_dir, "file1.txt")
            file2 = os.path.join(source_dir, "file2.txt")
            with open(file1, "w") as f:
                f.write("content1")
            with open(file2, "w") as f:
                f.write("content2")

            result, file_count = consolidate_metadata(
                {"source": ["file1.txt", "file2.txt"]}, tmpdir
            )

            assert os.path.exists(os.path.join(result, "file1.txt"))
            assert os.path.exists(os.path.join(result, "file2.txt"))
            assert file_count == 2
            clean_temp_directory(result)

    def test_consolidate_with_absolute_path(self):
        """Test consolidating with absolute path."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content")

            abs_path = os.path.abspath(source_dir)
            result, file_count = consolidate_metadata(abs_path, tmpdir)

            assert os.path.exists(os.path.join(result, "file.txt"))
            assert file_count == 1
            clean_temp_directory(result)

    def test_consolidate_invalid_type_raises_error(self):
        """Test consolidating with invalid type raises TypeError or ValueError."""
        # TypeGuard will catch type errors before ValueError
        with pytest.raises((TypeError, ValueError)):
            consolidate_metadata(12345)  # Invalid type

    def test_consolidate_invalid_pattern_type_raises_error(self):
        """Test consolidating with invalid pattern type raises TypeError or ValueError."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            # TypeGuard will catch type errors before ValueError
            with pytest.raises((TypeError, ValueError)):
                consolidate_metadata({"source": 12345}, tmpdir)

    def test_consolidate_nonexistent_directory_raises_error(self):
        """Test consolidating non-existent directory raises ValueError."""
        with pytest.raises(ValueError):
            consolidate_metadata("nonexistent", "/tmp")

    def test_consolidate_cleans_up_on_error(self):
        """Test that temp directory is cleaned up on error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                consolidate_metadata("nonexistent", tmpdir)
            except ValueError:
                pass

            # Temp directory should be cleaned up
            # We can't directly check, but if it wasn't cleaned up,
            # we'd have issues with temp directory accumulation

    def test_consolidate_dict_with_star_pattern(self):
        """Test consolidating with '*' pattern."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            src_subdir = os.path.join(source_dir, "src")
            os.makedirs(src_subdir)

            with open(os.path.join(src_subdir, "file.txt"), "w") as f:
                f.write("content")

            result, file_count = consolidate_metadata({"source": "*"}, tmpdir)

            assert os.path.exists(os.path.join(result, "src", "file.txt"))
            assert file_count == 1
            clean_temp_directory(result)


class TestPrintDirectoryTree:
    """Test cases for print_directory_tree function."""

    def test_print_tree_without_logger(self, capsys):
        """Test printing directory tree without logger."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("content1")
            os.makedirs(os.path.join(tmpdir, "subdir"))
            with open(os.path.join(tmpdir, "subdir", "file2.txt"), "w") as f:
                f.write("content2")

            print_directory_tree(tmpdir)

            captured = capsys.readouterr()
            assert "file1.txt" in captured.out
            assert "subdir" in captured.out

    def test_print_tree_with_logger(self):
        """Test printing directory tree with logger."""
        import logging

        with tempfile.TemporaryDirectory() as tmpdir:
            with open(os.path.join(tmpdir, "file.txt"), "w") as f:
                f.write("content")

            logger = logging.getLogger("test")
            with mock.patch.object(logger, "info") as mock_info:
                print_directory_tree(tmpdir, logger=logger)

                # Verify logger.info was called
                assert mock_info.called
                call_args = [args[0] for args, _ in mock_info.call_args_list]
                assert any("file.txt" in msg for msg in call_args)

    def test_print_tree_respects_max_depth(self, capsys):
        """Test that print_tree respects max_depth parameter."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create nested structure
            current = tmpdir
            for i in range(5):
                current = os.path.join(current, f"level{i}")
                os.makedirs(current)

            print_directory_tree(tmpdir, max_depth=2)

            captured = capsys.readouterr()
            # Should only show first 2 levels
            assert "level0" in captured.out

    def test_print_tree_handles_permission_error(self, capsys):
        """Test that print_tree handles PermissionError gracefully."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Mock os.listdir to raise PermissionError
            with mock.patch("os.listdir", side_effect=PermissionError("Access denied")):
                print_directory_tree(tmpdir)
                # Should not raise exception


class TestConsolidateUnpackagedMetadataTask:
    """Test cases for ConsolidateUnpackagedMetadata task class."""

    def setup_method(self):
        """Set up test fixtures."""
        self.temp_repo_root = tempfile.mkdtemp()
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config,
            config={"noyaml": True, "project": {"package": {}}},
            repo_info={"root": self.temp_repo_root},
        )

    def teardown_method(self):
        """Clean up test fixtures."""
        if os.path.exists(self.temp_repo_root):
            shutil.rmtree(self.temp_repo_root)

    def test_task_with_string_metadata_path(self):
        """Test task with string metadata path."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        with open(os.path.join(source_dir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        # Task stores result in task.result (from _run_task return value)
        result_path = task.result

        assert result_path is not None
        assert isinstance(result_path, str)
        assert os.path.exists(result_path)
        assert os.path.exists(os.path.join(result_path, "file.txt"))
        clean_temp_directory(result_path)

    def test_task_with_list_metadata_path(self):
        """Test task with list metadata path."""
        dir1 = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        dir2 = os.path.join(self.temp_repo_root, "unpackaged", "post")
        os.makedirs(dir1)
        os.makedirs(dir2)

        with open(os.path.join(dir1, "file1.txt"), "w") as f:
            f.write("content1")
        with open(os.path.join(dir2, "file2.txt"), "w") as f:
            f.write("content2")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": ["unpackaged/pre", "unpackaged/post"]
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        result_path = task.result
        assert result_path is not None
        assert os.path.exists(os.path.join(result_path, "file1.txt"))
        assert os.path.exists(os.path.join(result_path, "file2.txt"))
        clean_temp_directory(result_path)

    def test_task_with_dict_metadata_path(self):
        """Test task with dict metadata path."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        test_file = os.path.join(source_dir, "test.txt")
        with open(test_file, "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": {"unpackaged/pre": "test.txt"}
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        result_path = task.result
        assert result_path is not None
        assert os.path.exists(os.path.join(result_path, "test.txt"))
        clean_temp_directory(result_path)

    def test_task_with_no_metadata_path(self):
        """Test task with no metadata path configured."""
        self.project_config.config["project"]["package"] = {}

        task_config = TaskConfig({})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        with mock.patch.object(task.logger, "warning") as mock_warning:
            task()

            mock_warning.assert_called_once()
            assert task.return_values["path"] is None

    def test_task_with_base_path_option(self):
        """Test task with custom base_path option."""
        custom_base = tempfile.mkdtemp()
        try:
            source_dir = os.path.join(custom_base, "source")
            os.makedirs(source_dir)

            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content")

            self.project_config.config["project"]["package"] = {
                "unpackaged_metadata_path": "source"
            }

            task_config = TaskConfig(
                {"options": {"base_path": custom_base, "keep_temp": True}}
            )
            task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

            task()

            assert task.result is not None
            assert os.path.exists(os.path.join(task.result, "file.txt"))
            clean_temp_directory(task.result)
        finally:
            shutil.rmtree(custom_base)

    def test_task_with_keep_temp_option(self):
        """Test task with keep_temp option enabled."""
        source_dir = os.path.join(self.project_config.repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        with open(os.path.join(source_dir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        assert task.result is not None
        # Directory should still exist since keep_temp=True
        assert os.path.exists(task.result)
        clean_temp_directory(task.result)

    def test_task_logs_consolidation_info(self):
        """Test that task logs consolidation information."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        with mock.patch.object(task.logger, "info") as mock_info:
            task()

            # Check that info was called with consolidation messages
            call_args = [call[0][0] for call in mock_info.call_args_list]
            assert any("Consolidating unpackaged metadata" in msg for msg in call_args)
            assert any("Found" in msg for msg in call_args)


class TestWindowsLinuxCompatibility:
    """Test cases for Windows/Linux path compatibility."""

    def test_path_separators_handled_correctly(self):
        """Test that path separators are handled correctly on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Use os.path.join which handles platform differences
            source_dir = os.path.join(tmpdir, "source", "subdir")
            os.makedirs(source_dir)

            test_file = os.path.join(source_dir, "file.txt")
            with open(test_file, "w") as f:
                f.write("content")

            # Test with relative pattern from source directory
            pattern = "file.txt"
            result, file_count = consolidate_metadata(
                {os.path.join(tmpdir, "source", "subdir"): pattern}, tmpdir
            )

            assert os.path.exists(result)
            assert file_count == 1
            assert os.path.exists(os.path.join(result, "file.txt"))
            clean_temp_directory(result)

    def test_absolute_paths_work_on_both_platforms(self):
        """Test that absolute paths work on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content")

            abs_path = os.path.abspath(source_dir)
            result, file_count = consolidate_metadata(abs_path, tmpdir)

            assert os.path.exists(os.path.join(result, "file.txt"))
            assert file_count == 1
            clean_temp_directory(result)

    def test_relative_paths_work_on_both_platforms(self):
        """Test that relative paths work on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            original_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                source_dir = "source"
                os.makedirs(source_dir)

                with open(os.path.join(source_dir, "file.txt"), "w") as f:
                    f.write("content")

                result, file_count = consolidate_metadata(
                    source_dir, tmpdir, logger=None
                )

                assert os.path.exists(os.path.join(result, "file.txt"))
                assert file_count == 1
                clean_temp_directory(result)
            finally:
                os.chdir(original_cwd)

    def test_nested_paths_work_on_both_platforms(self):
        """Test that deeply nested paths work on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create deeply nested structure
            nested_path = os.path.join(tmpdir, "level1", "level2", "level3")
            os.makedirs(nested_path)

            test_file = os.path.join(nested_path, "file.txt")
            with open(test_file, "w") as f:
                f.write("content")

            # Test merging nested directories
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(dest_dir)

            merge_directory_contents(os.path.join(tmpdir, "level1"), dest_dir)

            assert os.path.exists(
                os.path.join(dest_dir, "level2", "level3", "file.txt")
            )

    def test_path_with_spaces_works_on_both_platforms(self):
        """Test that paths with spaces work on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source with spaces")
            dest_dir = os.path.join(tmpdir, "dest with spaces")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            test_file = os.path.join(source_dir, "file with spaces.txt")
            with open(test_file, "w") as f:
                f.write("content")

            copy_directory_contents(source_dir, dest_dir)

            assert os.path.exists(os.path.join(dest_dir, "file with spaces.txt"))


class TestMergeDirectoryContentsEdgeCases:
    """Additional edge case tests for merge_directory_contents."""

    def test_merge_with_file_conflict_no_overwrite(self):
        """Test merge when file exists in destination but overwrite=False."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create file in destination
            dest_file = os.path.join(dest_dir, "test.txt")
            with open(dest_file, "w") as f:
                f.write("dest content")

            # Create file in source with different content
            src_file = os.path.join(src_dir, "test.txt")
            with open(src_file, "w") as f:
                f.write("src content")

            merge_directory_contents(src_dir, dest_dir, overwrite=False)

            # Should overwrite even without overwrite flag (default behavior)
            with open(dest_file, "r") as f:
                assert f.read() == "src content"

    def test_merge_directory_replaces_file_with_overwrite(self):
        """Test merge when source has directory but dest has file with overwrite=True."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create file in destination
            dest_file = os.path.join(dest_dir, "item")
            with open(dest_file, "w") as f:
                f.write("file")

            # Create directory with same name in source
            src_subdir = os.path.join(src_dir, "item")
            os.makedirs(src_subdir)
            with open(os.path.join(src_subdir, "file.txt"), "w") as f:
                f.write("content")

            merge_directory_contents(src_dir, dest_dir, overwrite=True)

            # Should be replaced with directory
            assert os.path.isdir(os.path.join(dest_dir, "item"))
            assert os.path.exists(os.path.join(dest_dir, "item", "file.txt"))

    def test_merge_file_replaces_directory_with_overwrite(self):
        """Test merge when source has file but dest has directory with overwrite=True."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create directory in destination
            dest_subdir = os.path.join(dest_dir, "item")
            os.makedirs(dest_subdir)

            # Create file with same name in source
            src_file = os.path.join(src_dir, "item")
            with open(src_file, "w") as f:
                f.write("file content")

            merge_directory_contents(src_dir, dest_dir, overwrite=True)

            # Should be replaced with file
            assert os.path.isfile(os.path.join(dest_dir, "item"))
            with open(os.path.join(dest_dir, "item"), "r") as f:
                assert f.read() == "file content"

    def test_merge_empty_source_directory(self):
        """Test merging empty source directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            merge_directory_contents(src_dir, dest_dir)
            # Should not raise error
            assert os.path.exists(dest_dir)


class TestCopyItemToDestinationEdgeCases:
    """Additional edge case tests for copy_item_to_destination."""

    def test_copy_file_when_dest_is_directory_no_overwrite(self):
        """Test copying file when destination is directory without overwrite."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_file = os.path.join(tmpdir, "source.txt")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(dest_dir)

            with open(src_file, "w") as f:
                f.write("file content")

            # Should raise error or handle gracefully
            try:
                copy_item_to_destination(src_file, dest_dir, overwrite=False)
                # If no error, file should be copied into directory
                assert os.path.exists(os.path.join(dest_dir, "source.txt"))
            except (OSError, shutil.Error):
                # Some systems may raise error
                pass

    def test_copy_directory_when_dest_is_file_no_overwrite(self):
        """Test copying directory when destination is file without overwrite."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_file = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)

            src_file = os.path.join(src_dir, "file.txt")
            with open(src_file, "w") as f:
                f.write("content")

            with open(dest_file, "w") as f:
                f.write("file")

            # Code removes the file and copies directory
            copy_item_to_destination(src_dir, dest_file, overwrite=False)

            # Should be replaced with directory
            assert os.path.isdir(dest_file)
            assert os.path.exists(os.path.join(dest_file, "file.txt"))

    def test_copy_to_nonexistent_parent_directory(self):
        """Test copying to path with nonexistent parent directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_file = os.path.join(tmpdir, "source.txt")
            dest_file = os.path.join(tmpdir, "nonexistent", "dest.txt")

            with open(src_file, "w") as f:
                f.write("content")

            # Should raise error
            with pytest.raises((OSError, FileNotFoundError)):
                copy_item_to_destination(src_file, dest_file)


class TestCopyDirectoryContentsEdgeCases:
    """Additional edge case tests for copy_directory_contents."""

    def test_copy_empty_directory(self):
        """Test copying empty directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            copy_directory_contents(src_dir, dest_dir)
            # Should not raise error
            assert os.path.exists(dest_dir)

    def test_extract_src_with_overwrite(self):
        """Test extract_src with overwrite flag."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create src subdirectory
            src_subdir = os.path.join(src_dir, "src")
            os.makedirs(src_subdir)
            with open(os.path.join(src_subdir, "file.txt"), "w") as f:
                f.write("content")

            # Create conflicting file in dest
            dest_file = os.path.join(dest_dir, "src", "file.txt")
            os.makedirs(os.path.dirname(dest_file), exist_ok=True)
            with open(dest_file, "w") as f:
                f.write("old")

            copy_directory_contents(src_dir, dest_dir, extract_src=True, overwrite=True)

            # Should overwrite
            with open(dest_file, "r") as f:
                assert f.read() == "content"

    def test_extract_src_with_nested_src_directories(self):
        """Test extract_src with nested src directories."""
        with tempfile.TemporaryDirectory() as tmpdir:
            src_dir = os.path.join(tmpdir, "src")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(src_dir)
            os.makedirs(dest_dir)

            # Create nested src structure
            nested_src = os.path.join(src_dir, "src", "nested", "src")
            os.makedirs(nested_src)
            with open(os.path.join(nested_src, "file.txt"), "w") as f:
                f.write("content")

            copy_directory_contents(src_dir, dest_dir, extract_src=True)

            # Should extract only the top-level src
            assert os.path.exists(
                os.path.join(dest_dir, "src", "nested", "src", "file.txt")
            )


class TestResolveFilePatternEdgeCases:
    """Additional edge case tests for resolve_file_pattern."""

    def test_resolve_pattern_with_dot_files(self):
        """Test resolving pattern matching dot files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            dot_file = os.path.join(tmpdir, ".hidden")
            with open(dot_file, "w") as f:
                f.write("content")

            result = resolve_file_pattern(".hidden", tmpdir)
            assert len(result) == 1
            assert result[0] == dot_file

    def test_resolve_pattern_with_multiple_extensions(self):
        """Test resolving pattern with multiple file extensions."""
        with tempfile.TemporaryDirectory() as tmpdir:
            for ext in [".txt", ".xml", ".json"]:
                with open(os.path.join(tmpdir, f"file{ext}"), "w") as f:
                    f.write("content")

            result = resolve_file_pattern("*.*", tmpdir)
            assert len(result) >= 3

    def test_resolve_pattern_with_question_mark(self):
        """Test resolving pattern with question mark wildcard."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("content2")
            with open(os.path.join(tmpdir, "file10.txt"), "w") as f:
                f.write("content10")

            result = resolve_file_pattern("file?.txt", tmpdir)
            assert len(result) == 2  # file1.txt and file2.txt, not file10.txt

    def test_resolve_pattern_with_character_class(self):
        """Test resolving pattern with character class."""
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(os.path.join(tmpdir, "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(tmpdir, "file2.txt"), "w") as f:
                f.write("content2")
            with open(os.path.join(tmpdir, "filea.txt"), "w") as f:
                f.write("contenta")

            result = resolve_file_pattern("file[12].txt", tmpdir)
            assert len(result) == 2


class TestCopyMatchedFilesEdgeCases:
    """Additional edge case tests for copy_matched_files."""

    def test_copy_empty_file_list(self):
        """Test copying empty list of files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            copy_matched_files([], source_dir, dest_dir)
            # Should not raise error
            assert os.path.exists(dest_dir)

    def test_copy_file_at_root_level(self):
        """Test copying file at root level of source directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            source_file = os.path.join(source_dir, "root.txt")
            with open(source_file, "w") as f:
                f.write("content")

            copy_matched_files([source_file], source_dir, dest_dir)

            dest_file = os.path.join(dest_dir, "root.txt")
            assert os.path.exists(dest_file)

    def test_copy_files_with_same_name_different_paths(self):
        """Test copying files with same name from different paths."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            file1 = os.path.join(source_dir, "dir1", "file.txt")
            file2 = os.path.join(source_dir, "dir2", "file.txt")
            os.makedirs(os.path.dirname(file1))
            os.makedirs(os.path.dirname(file2))

            with open(file1, "w") as f:
                f.write("content1")
            with open(file2, "w") as f:
                f.write("content2")

            copy_matched_files([file1, file2], source_dir, dest_dir)

            assert os.path.exists(os.path.join(dest_dir, "dir1", "file.txt"))
            assert os.path.exists(os.path.join(dest_dir, "dir2", "file.txt"))


class TestConsolidateMetadataEdgeCases:
    """Additional edge case tests for consolidate_metadata."""

    def test_consolidate_dict_with_empty_list_patterns(self):
        """Test consolidating with empty list of patterns."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            result, file_count = consolidate_metadata({"source": []}, tmpdir)
            # Should create temp dir but no files
            assert os.path.exists(result)
            assert file_count == 0
            clean_temp_directory(result)

    def test_consolidate_dict_with_multiple_wildcards(self):
        """Test consolidating with multiple directories using wildcards."""
        with tempfile.TemporaryDirectory() as tmpdir:
            dir1 = os.path.join(tmpdir, "dir1")
            dir2 = os.path.join(tmpdir, "dir2")
            os.makedirs(os.path.join(dir1, "src"))
            os.makedirs(os.path.join(dir2, "src"))

            with open(os.path.join(dir1, "src", "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(dir2, "src", "file2.txt"), "w") as f:
                f.write("content2")

            result, file_count = consolidate_metadata(
                {"dir1": "*", "dir2": "*"}, tmpdir
            )

            assert os.path.exists(os.path.join(result, "src", "file1.txt"))
            assert os.path.exists(os.path.join(result, "src", "file2.txt"))
            assert file_count == 2
            clean_temp_directory(result)

    def test_consolidate_list_with_empty_strings(self):
        """Test consolidating list with empty strings."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content")

            # Empty string resolves to base_path, which is valid
            # So it will copy base_path contents, then source contents
            result, file_count = consolidate_metadata(["", "source"], tmpdir)

            # Should succeed and consolidate both
            assert os.path.exists(result)
            # Source file should be present
            assert os.path.exists(os.path.join(result, "file.txt"))
            assert file_count == 2
            clean_temp_directory(result)

    def test_consolidate_verifies_cleanup_on_error(self):
        """Test that temp directory is cleaned up on validation error."""
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_dirs_before = len(
                [
                    d
                    for d in os.listdir(tempfile.gettempdir())
                    if d.startswith("metadata_consolidate_")
                ]
            )

            try:
                consolidate_metadata("nonexistent", tmpdir)
            except ValueError:
                pass

            # Verify no new temp directories were left behind
            temp_dirs_after = len(
                [
                    d
                    for d in os.listdir(tempfile.gettempdir())
                    if d.startswith("metadata_consolidate_")
                ]
            )
            # Should be same or less (cleanup happened)
            assert temp_dirs_after <= temp_dirs_before + 1

    def test_consolidate_dict_with_nonexistent_pattern(self):
        """Test consolidating with pattern that doesn't match files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            os.makedirs(source_dir)

            result, file_count = consolidate_metadata(
                {"source": "nonexistent.txt"}, tmpdir
            )
            assert file_count == 0
            assert result is not None

    def test_consolidate_with_base_path_none(self):
        """Test consolidate with base_path explicitly None."""
        with tempfile.TemporaryDirectory() as tmpdir:
            original_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                source_dir = "source"
                os.makedirs(source_dir)

                with open(os.path.join(source_dir, "file.txt"), "w") as f:
                    f.write("content")

                result, file_count = consolidate_metadata(source_dir, base_path=None)
                assert os.path.exists(os.path.join(result, "file.txt"))
                assert file_count == 1
                clean_temp_directory(result)
            finally:
                os.chdir(original_cwd)


class TestConsolidateUnpackagedMetadataTaskEdgeCases:
    """Additional edge case tests for ConsolidateUnpackagedMetadata task."""

    def setup_method(self):
        """Set up test fixtures."""
        self.temp_repo_root = tempfile.mkdtemp()
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config,
            config={"noyaml": True, "project": {"package": {}}},
            repo_info={"root": self.temp_repo_root},
        )

    def teardown_method(self):
        """Clean up test fixtures."""
        if os.path.exists(self.temp_repo_root):
            shutil.rmtree(self.temp_repo_root)

    def test_task_return_value_structure(self):
        """Test that task returns correct value structure."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        with open(os.path.join(source_dir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        # Task stores result in task.result
        assert isinstance(task.result, str)
        assert os.path.exists(task.result)
        # With keep_temp=True, directory should still exist
        clean_temp_directory(task.result)

    def test_task_with_error_during_consolidation(self):
        """Test task behavior when consolidation raises error."""
        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "nonexistent/path"
        }

        task_config = TaskConfig({})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        with pytest.raises(ValueError):
            task()

    def test_task_cleans_up_temp_when_keep_temp_false(self):
        """Test that task cleans up temp directory when keep_temp=False."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        with open(os.path.join(source_dir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({"options": {"keep_temp": False}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        # Directory should be cleaned up (but result still contains the path)
        result_path = task.result
        assert result_path is not None
        assert not os.path.exists(result_path)

    def test_task_with_dict_metadata_path_wildcard(self):
        """Test task with dict metadata path using wildcard."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        src_subdir = os.path.join(source_dir, "src")
        os.makedirs(src_subdir)

        with open(os.path.join(src_subdir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": {"unpackaged/pre": "*.*"}
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        result_path = task.result
        assert result_path is not None
        assert os.path.exists(os.path.join(result_path, "src", "file.txt"))
        clean_temp_directory(result_path)

    def test_task_with_list_metadata_path_multiple_dirs(self):
        """Test task with list metadata path containing multiple directories."""
        dir1 = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        dir2 = os.path.join(self.temp_repo_root, "unpackaged", "post")
        dir3 = os.path.join(self.temp_repo_root, "unpackaged", "default")
        os.makedirs(dir1)
        os.makedirs(dir2)
        os.makedirs(dir3)

        with open(os.path.join(dir1, "file1.txt"), "w") as f:
            f.write("content1")
        with open(os.path.join(dir2, "file2.txt"), "w") as f:
            f.write("content2")
        with open(os.path.join(dir3, "file3.txt"), "w") as f:
            f.write("content3")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": [
                "unpackaged/pre",
                "unpackaged/post",
                "unpackaged/default",
            ]
        }

        task_config = TaskConfig({"options": {"keep_temp": True}})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        task()

        result_path = task.result
        assert result_path is not None
        assert os.path.exists(os.path.join(result_path, "file1.txt"))
        assert os.path.exists(os.path.join(result_path, "file2.txt"))
        assert os.path.exists(os.path.join(result_path, "file3.txt"))
        clean_temp_directory(result_path)

    def test_task_logs_tree_structure(self):
        """Test that task logs directory tree structure."""
        source_dir = os.path.join(self.temp_repo_root, "unpackaged", "pre")
        os.makedirs(source_dir)

        with open(os.path.join(source_dir, "file.txt"), "w") as f:
            f.write("content")

        self.project_config.config["project"]["package"] = {
            "unpackaged_metadata_path": "unpackaged/pre"
        }

        task_config = TaskConfig({})
        task = ConsolidateUnpackagedMetadata(self.project_config, task_config, None)

        with mock.patch.object(task.logger, "info") as mock_info:
            task()

            # Check that tree printing was called
            call_args = [call[0][0] for call in mock_info.call_args_list]
            # Should have tree structure calls (containing tree characters)
            tree_calls = [msg for msg in call_args if "├──" in msg or "└──" in msg]
            assert len(tree_calls) > 0


class TestWindowsLinuxCompatibilityExtended:
    """Extended Windows/Linux compatibility tests."""

    def test_paths_with_special_characters(self):
        """Test paths with special characters that work on both platforms."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Use characters that are valid on both platforms
            source_dir = os.path.join(tmpdir, "source-dir_123")
            dest_dir = os.path.join(tmpdir, "dest-dir_456")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            test_file = os.path.join(source_dir, "file-name_123.txt")
            with open(test_file, "w") as f:
                f.write("content")

            copy_directory_contents(source_dir, dest_dir)

            assert os.path.exists(os.path.join(dest_dir, "file-name_123.txt"))

    def test_paths_with_unicode_characters(self):
        """Test paths with unicode characters."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source_测试")
            dest_dir = os.path.join(tmpdir, "dest_テスト")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            test_file = os.path.join(source_dir, "file_文件.txt")
            with open(test_file, "w", encoding="utf-8") as f:
                f.write("content")

            copy_directory_contents(source_dir, dest_dir)

            assert os.path.exists(os.path.join(dest_dir, "file_文件.txt"))

    def test_relative_paths_with_dot_dot(self):
        """Test relative paths with .. components."""
        with tempfile.TemporaryDirectory() as tmpdir:
            nested_dir = os.path.join(tmpdir, "level1", "level2")
            os.makedirs(nested_dir)

            test_file = os.path.join(nested_dir, "file.txt")
            with open(test_file, "w") as f:
                f.write("content")

            # Use relative path with .. - resolve to absolute first
            original_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                # Use absolute path or resolve the relative path properly
                abs_nested = os.path.abspath(nested_dir)
                result, file_count = consolidate_metadata(abs_nested, tmpdir)
                assert os.path.exists(os.path.join(result, "file.txt"))
                assert file_count == 1
                clean_temp_directory(result)
            finally:
                os.chdir(original_cwd)

    def test_symlink_handling(self):
        """Test handling of symlinks (if supported on platform)."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            # Create a regular file
            real_file = os.path.join(tmpdir, "real.txt")
            with open(real_file, "w") as f:
                f.write("content")

            # Try to create symlink (may not work on Windows without admin)
            symlink_file = os.path.join(source_dir, "link.txt")
            try:
                os.symlink(real_file, symlink_file)

                copy_directory_contents(source_dir, dest_dir)

                # Symlink should be copied (as file or link depending on platform)
                assert os.path.exists(os.path.join(dest_dir, "link.txt"))
            except (OSError, NotImplementedError):
                # Symlinks not supported on this platform
                pass

    def test_long_path_names(self):
        """Test handling of long path names."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create nested structure with long names
            long_path = tmpdir
            for i in range(5):
                long_path = os.path.join(long_path, f"very_long_directory_name_{i}" * 5)
                os.makedirs(long_path)

            test_file = os.path.join(long_path, "file.txt")
            with open(test_file, "w") as f:
                f.write("content")

            # Test that we can still copy
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(dest_dir)

            copy_directory_contents(
                os.path.join(tmpdir, "very_long_directory_name_0" * 5), dest_dir
            )

            # Should handle long paths
            assert os.path.exists(dest_dir)

    def test_case_sensitivity_handling(self):
        """Test handling of case-sensitive vs case-insensitive file systems."""
        with tempfile.TemporaryDirectory() as tmpdir:
            source_dir = os.path.join(tmpdir, "source")
            dest_dir = os.path.join(tmpdir, "dest")
            os.makedirs(source_dir)
            os.makedirs(dest_dir)

            # Create files with different cases
            with open(os.path.join(source_dir, "File.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(source_dir, "file.txt"), "w") as f:
                f.write("content2")

            copy_directory_contents(source_dir, dest_dir)

            # Both should exist (or overwrite depending on platform)
            assert os.path.exists(os.path.join(dest_dir, "File.txt")) or os.path.exists(
                os.path.join(dest_dir, "file.txt")
            )



import datetime
import os
import re

import pytest
import pytz

from cumulusci.core.exceptions import ConfigMergeError, TaskOptionsError
from cumulusci.utils import temporary_dir, touch

from .. import utils


class TestUtils:
    def test_parse_datetime(self):
        dt = utils.parse_datetime("2018-07-30", "%Y-%m-%d")
        assert dt == datetime.datetime(2018, 7, 30, 0, 0, 0, 0, pytz.UTC)

    def test_process_bool_arg(self):
        for arg in (True, "True", "true", "1"):
            assert utils.process_bool_arg(arg)

        for arg in (False, "False", "false", "0"):
            assert not utils.process_bool_arg(arg)

        import warnings

        with warnings.catch_warnings(record=True):
            assert utils.process_bool_arg(None) is False

        with pytest.raises(TypeError):
            utils.process_bool_arg(datetime.datetime.now())

        with pytest.raises(TypeError):
            utils.process_bool_arg("xyzzy")

    def test_process_list_arg(self):
        assert [1, 2] == utils.process_list_arg([1, 2])
        assert ["a", "b"] == utils.process_list_arg("a, b")
        assert utils.process_list_arg(None) is None

    def test_process_glob_list_arg(self):
        with temporary_dir():
            touch("foo.py")
            touch("bar.robot")

            # Expect passing arg as list works.
            assert ["foo.py", "bar.robot"] == utils.process_glob_list_arg(
                ["foo.py", "bar.robot"]
            )

            # Falsy arg should return an empty list
            assert [] == utils.process_glob_list_arg(None)
            assert [] == utils.process_glob_list_arg("")
            assert [] == utils.process_glob_list_arg([])

            # Expect output to be in order given
            assert ["foo.py", "bar.robot"] == utils.process_glob_list_arg(
                "foo.py, bar.robot"
            )

            # Expect sorted output of glob results
            assert ["bar.robot", "foo.py"] == utils.process_glob_list_arg("*")

            # Patterns that don't match any files
            assert ["*.bar", "x.y.z"] == utils.process_glob_list_arg("*.bar, x.y.z")

            # Recursive
            os.mkdir("subdir")
            filename = os.path.join("subdir", "baz.resource")
            touch(filename)
            assert [filename] == utils.process_glob_list_arg("**/*.resource")

    def test_decode_to_unicode(self):
        assert utils.decode_to_unicode(b"\xfc") == "\xfc"
        assert utils.decode_to_unicode("\u2603") == "\u2603"
        assert utils.decode_to_unicode(None) is None

    json_test_cases = [
        ({1, 2, 3}, [1, 2, 3]),
        ("abc", "abc"),
        ([1, 2, 3], [1, 2, 3]),
        (b"bytes", str(b"bytes")),
    ]

    @pytest.mark.parametrize("input, expected_return", json_test_cases)
    def test_make_jsonable(self, input, expected_return):
        assert utils.make_jsonable(input) == expected_return


class TestDictMerger:
    """some stuff that didnt get covered by usual usage"""

    def test_merge_into_list(self):
        combo = utils.dictmerge([1, 2], 3)
        assert combo == [1, 2, 3]

    def test_cant_merge_into_dict(self):
        with pytest.raises(ConfigMergeError):
            utils.dictmerge({"a": "b"}, 2)

    def test_cant_merge_nonsense(self):
        with pytest.raises(ConfigMergeError):
            utils.dictmerge(pytz, 2)


class TestProcessListOfPairsDictArg:
    def test_process_list_of_pairs_dict_arg__already_dict(self):
        expected_dict = {"foo": "bar"}
        actual_dict = utils.process_list_of_pairs_dict_arg(expected_dict)
        assert actual_dict is expected_dict

    def test_process_list_of_pairs_dict_arg__valid_values(self):
        valid_values = "foo:bar,baz:boo"
        actual_dict = utils.process_list_of_pairs_dict_arg(valid_values)
        assert actual_dict == {"foo": "bar", "baz": "boo"}

    def test_process_list_of_pairs_dict_arg__uri_values(self):
        uri_value = "companyWebsite:https://www.salesforce.org:8080"
        actual_dict = utils.process_list_of_pairs_dict_arg(uri_value)
        assert actual_dict == {"companyWebsite": "https://www.salesforce.org:8080"}

    def test_process_list_of_pairs_dict_arg__not_dict_or_string(self):
        unsupported = ("foo", "bar")
        error_message = re.escape(
            f"Arg is not a dict or string ({type(unsupported)}): {unsupported}"
        )
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(unsupported)

    def test_process_list_of_pairs_dict_arg__not_name_value_pair(self):
        not_pair = "foo:bar,baz"
        error_message = re.escape("Var is not a name/value pair: baz")
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(not_pair)

    def test_process_list_of_pairs_dict_arg__duplicate_value(self):
        duplicate = "foo:bar,foo:baz"
        error_message = re.escape("Var specified twice: foo")
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(duplicate)


class TestDeepMergePlugins:
    """Test the deep_merge_plugins function"""

    def test_deep_merge_plugins_remote_takes_precedence(self):
        """Test that remote plugins take precedence over project plugins"""
        remote_plugins = {"plugin1": {"setting": "remote_value"}}
        project_plugins = {"plugin1": {"setting": "project_value"}}

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        assert result == {"plugin1": {"setting": "remote_value"}}

    def test_deep_merge_plugins_project_provides_defaults(self):
        """Test that project plugins provide defaults for missing keys"""
        remote_plugins = {"plugin1": {"setting1": "remote_value"}}
        project_plugins = {"plugin1": {"setting2": "project_value"}}

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        assert result == {
            "plugin1": {"setting1": "remote_value", "setting2": "project_value"}
        }

    def test_deep_merge_plugins_missing_keys_from_project(self):
        """Test that missing top-level keys are added from project plugins"""
        remote_plugins = {"plugin1": {"setting": "remote_value"}}
        project_plugins = {"plugin2": {"setting": "project_value"}}

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        assert result == {
            "plugin1": {"setting": "remote_value"},
            "plugin2": {"setting": "project_value"},
        }

    def test_deep_merge_plugins_recursive_merge(self):
        """Test recursive merging of nested dictionaries"""
        remote_plugins = {
            "plugin1": {
                "nested": {"setting1": "remote_value", "setting2": "remote_value2"}
            }
        }
        project_plugins = {
            "plugin1": {
                "nested": {
                    "setting2": "project_value2",  # Should be overridden
                    "setting3": "project_value3",  # Should be added
                }
            }
        }

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        assert result == {
            "plugin1": {
                "nested": {
                    "setting1": "remote_value",
                    "setting2": "remote_value2",
                    "setting3": "project_value3",
                }
            }
        }

    def test_deep_merge_plugins_non_dict_inputs(self):
        """Test that non-dict inputs return the remote plugins unchanged"""
        remote_plugins = {"plugin1": {"setting": "value"}}

        # Test with non-dict project_plugins
        result = utils.deep_merge_plugins(remote_plugins, "not_a_dict")
        assert result == remote_plugins

        # Test with non-dict remote_plugins
        result = utils.deep_merge_plugins(
            "not_a_dict", {"plugin1": {"setting": "value"}}
        )
        assert result == "not_a_dict"

        # Test with both non-dict
        result = utils.deep_merge_plugins("remote", "project")
        assert result == "remote"

    def test_deep_merge_plugins_type_mismatch(self):
        """Test that type mismatches preserve remote values"""
        remote_plugins = {"plugin1": {"setting": "string_value"}}
        project_plugins = {"plugin1": {"setting": {"nested": "dict_value"}}}

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        # Remote takes precedence when types don't match
        assert result == {"plugin1": {"setting": "string_value"}}

    def test_deep_merge_plugins_deep_copy(self):
        """Test that deep copy is used to avoid modifying original data"""
        remote_plugins = {"plugin1": {"setting": "remote_value"}}
        project_plugins = {"plugin2": {"nested": {"setting": "project_value"}}}

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        # Modify the result to check if originals are affected
        result["plugin2"]["nested"]["setting"] = "modified_value"

        # Original should remain unchanged
        assert project_plugins["plugin2"]["nested"]["setting"] == "project_value"

    def test_deep_merge_plugins_empty_inputs(self):
        """Test with empty dictionaries"""
        # Empty remote, non-empty project
        result = utils.deep_merge_plugins({}, {"plugin1": {"setting": "value"}})
        assert result == {"plugin1": {"setting": "value"}}

        # Non-empty remote, empty project
        result = utils.deep_merge_plugins({"plugin1": {"setting": "value"}}, {})
        assert result == {"plugin1": {"setting": "value"}}

        # Both empty
        result = utils.deep_merge_plugins({}, {})
        assert result == {}

    def test_deep_merge_plugins_complex_nested_structure(self):
        """Test with complex nested structure"""
        remote_plugins = {
            "plugin1": {"level1": {"level2": {"remote_setting": "remote_value"}}}
        }
        project_plugins = {
            "plugin1": {
                "level1": {
                    "level2": {"project_setting": "project_value"},
                    "project_level2": "project_value2",
                },
                "project_level1": "project_value3",
            }
        }

        result = utils.deep_merge_plugins(remote_plugins, project_plugins)

        expected = {
            "plugin1": {
                "level1": {
                    "level2": {
                        "remote_setting": "remote_value",
                        "project_setting": "project_value",
                    },
                    "project_level2": "project_value2",
                },
                "project_level1": "project_value3",
            }
        }

        assert result == expected



import pathlib
from typing import List, Optional, Union

from defusedxml.minidom import parseString
from pydantic.v1 import ValidationError

from cumulusci.cli.ui import CliTable
from cumulusci.core.dependencies.utils import TaskContext
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.sfdx import convert_sfdx_source
from cumulusci.core.source_transforms.transforms import (
    SourceTransform,
    SourceTransformList,
)
from cumulusci.core.utils import (
    determine_managed_mode,
    process_bool_arg,
    process_list_arg,
)
from cumulusci.salesforce_api.metadata import ApiDeploy, ApiRetrieveUnpackaged
from cumulusci.salesforce_api.package_zip import MetadataPackageZipBuilder
from cumulusci.salesforce_api.rest_deploy import RestDeploy
from cumulusci.tasks.metadata.package import process_common_components
from cumulusci.tasks.salesforce.BaseSalesforceMetadataApiTask import (
    BaseSalesforceMetadataApiTask,
)
from cumulusci.tasks.utility.copyContents import (
    clean_temp_directory,
    consolidate_metadata,
)
from cumulusci.utils.xml import metadata_tree


class Deploy(BaseSalesforceMetadataApiTask):
    api_class = ApiDeploy
    api_retrieve_unpackaged = ApiRetrieveUnpackaged
    task_options = {
        "path": {
            "description": "The path to the metadata source to be deployed",
            "required": True,
        },
        "unmanaged": {
            "description": "If True, changes namespace_inject to replace tokens with a blank string"
        },
        "namespace_inject": {
            "description": "If set, the namespace tokens in files and filenames are replaced with the namespace's prefix"
        },
        "namespace_strip": {
            "description": "If set, all namespace prefixes for the namespace specified are stripped from files and filenames"
        },
        "check_only": {
            "description": "If True, performs a test deployment (validation) of components without saving the components in the target org"
        },
        "collision_check": {
            "description": "If True, performs a collision check with metadata already present in the target org"
        },
        "test_level": {
            "description": "Specifies which tests are run as part of a deployment. Valid values: NoTestRun, RunLocalTests, RunAllTestsInOrg, RunSpecifiedTests."
        },
        "specified_tests": {
            "description": "Comma-separated list of test classes to run upon deployment. Applies only with test_level set to RunSpecifiedTests."
        },
        "static_resource_path": {
            "description": "The path where decompressed static resources are stored.  Any subdirectories found will be zipped and added to the staticresources directory of the build."
        },
        "namespaced_org": {
            "description": "If True, the tokens %%%NAMESPACED_ORG%%% and ___NAMESPACED_ORG___ will get replaced with the namespace.  The default is false causing those tokens to get stripped and replaced with an empty string.  Set this if deploying to a namespaced scratch org or packaging org."
        },
        "clean_meta_xml": {
            "description": "Defaults to True which strips the <packageVersions/> element from all meta.xml files.  The packageVersion element gets added automatically by the target org and is set to whatever version is installed in the org.  To disable this, set this option to False"
        },
        "transforms": {
            "description": "Apply source transforms before deploying. See the CumulusCI documentation for details on how to specify transforms."
        },
        "rest_deploy": {"description": "If True, deploy metadata using REST API"},
    }

    namespaces = {"sf": "http://soap.sforce.com/2006/04/metadata"}

    transforms: List[SourceTransform] = []

    def _init_options(self, kwargs):
        super(Deploy, self)._init_options(kwargs)

        self.check_only = process_bool_arg(self.options.get("check_only", False))
        self.test_level = self.options.get("test_level")
        if self.test_level and self.test_level not in [
            "NoTestRun",
            "RunLocalTests",
            "RunAllTestsInOrg",
            "RunSpecifiedTests",
        ]:
            raise TaskOptionsError(
                f"Specified test run level {self.test_level} is not valid."
            )

        self.specified_tests = process_list_arg(self.options.get("specified_tests", []))

        if bool(self.specified_tests) != (self.test_level == "RunSpecifiedTests"):
            raise TaskOptionsError(
                "The specified_tests option and test_level RunSpecifiedTests must be used together."
            )

        self.options["namespace_inject"] = (
            self.options.get("namespace_inject")
            or self.project_config.project__package__namespace
        )
        if "collision_check" not in self.options:
            self.options["collision_check"] = False

        if "transforms" in self.options:
            try:
                self.transforms = SourceTransformList.parse_obj(
                    self.options["transforms"]
                ).as_transforms()
            except ValidationError as e:
                raise TaskOptionsError(
                    "The transform spec is not valid. See CumulusCI documentation for details of how to specify transforms. "
                    f"The validation error was {str(e)}"
                )

        # Set class variable to true if rest_deploy is set to True
        self.rest_deploy = process_bool_arg(self.options.get("rest_deploy", False))

    def _get_api(self, path=None):
        if not path:
            path = self.options.get("path")

        package_zip = self._get_package_zip(path)

        if isinstance(package_zip, dict):
            self.logger.warning(
                "Deploy getting aborted due to collision of following components"
            )
            table_header_row = ["Type", "Component API Name"]
            table_data = [table_header_row]
            for type in package_zip.keys():
                for component_name in package_zip[type]:
                    table_data.append([type, component_name])
            table = CliTable(
                table_data,
            )
            table.echo()
            return None
        elif package_zip is not None:
            self.logger.info("Payload size: {} bytes".format(len(package_zip)))
        else:
            self.logger.warning("Deployment package is empty; skipping deployment.")
            return

        # If rest_deploy param is set, update api_class to be RestDeploy
        if self.rest_deploy:
            self.api_class = RestDeploy

        return self.api_class(
            self,
            package_zip,
            purge_on_delete=False,
            check_only=self.check_only,
            test_level=self.test_level,
            run_tests=self.specified_tests,
        )

    def _has_namespaced_package(self, ns: Optional[str]) -> bool:
        return determine_managed_mode(
            self.options, self.project_config, self.org_config
        )

    def _is_namespaced_org(self, ns: Optional[str]) -> bool:
        if "namespaced_org" in self.options:
            return process_bool_arg(self.options.get("namespaced_org", False))
        return bool(ns) and ns == self.org_config.namespace

    def _create_api_object(self, package_xml, api_version):
        api_retrieve_unpackaged_object = self.api_retrieve_unpackaged(
            self, package_xml, api_version
        )
        return api_retrieve_unpackaged_object

    def _collision_check(self, src_path):
        is_collision = False
        package_xml = open(f"{src_path}/package.xml", "r")
        source_xml_tree = metadata_tree.parse(f"{src_path}/package.xml")

        api_retrieve_unpackaged_response = self._create_api_object(
            package_xml.read(), source_xml_tree.version.text
        )

        xml_map = metadata_tree.parse_package_xml_types("name", source_xml_tree)

        messages = parseString(
            api_retrieve_unpackaged_response._get_response().content
        ).getElementsByTagName("messages")

        process_common_components(messages, xml_map)

        for type, api_names in xml_map.items():
            if len(api_names) != 0:
                is_collision = True
                break

        return is_collision, xml_map

    def _get_package_zip(self, path) -> Union[str, dict, None]:
        assert path, f"Path should be specified for {self.__class__.name}"
        if not pathlib.Path(path).exists():
            self.logger.warning(f"{path} not found.")
            return
        namespace = self.options["namespace_inject"]
        options = {
            **self.options,
            "clean_meta_xml": process_bool_arg(
                self.options.get("clean_meta_xml", True)
            ),
            "namespace_inject": namespace,
            "unmanaged": not self._has_namespaced_package(namespace),
            "namespaced_org": self._is_namespaced_org(namespace),
        }
        package_zip = None

        with convert_sfdx_source(path, None, self.logger) as src_path:
            ##############
            is_collision = False
            if "collision_check" in options and options["collision_check"]:
                is_collision, xml_map = self._collision_check(src_path)
            #############
            if not is_collision:
                context = TaskContext(self.org_config, self.project_config, self.logger)
                package_zip = MetadataPackageZipBuilder(
                    path=src_path,
                    context=context,
                    options=options,
                    transforms=self.transforms,
                )

                # If the package is empty, do nothing.
                if not package_zip.zf.namelist():
                    return
                return package_zip.as_base64()
            else:
                return xml_map

    def freeze(self, step):
        steps = super().freeze(step)
        for step in steps:
            if step["kind"] == "other":
                step["kind"] = "metadata"
        return steps


class DeployUnpackagedMetadata(Deploy):

    unpackaged_metadata_options = Deploy.task_options.copy()
    unpackaged_metadata_options.pop("path")
    task_options = {**unpackaged_metadata_options}

    def _init_options(self, kwargs):
        if self.project_config.project__package__unpackaged_metadata_path is None:
            return

        super(DeployUnpackagedMetadata, self)._init_options(kwargs)
        final_metadata_path, file_count = consolidate_metadata(
            self.project_config.project__package__unpackaged_metadata_path,
            self.project_config.repo_root,
        )
        if file_count == 0:
            self.logger.warning(
                "No files found in the unpackaged metadata path. Skipping metadata deployment task."
            )
            return
        self.options["path"] = final_metadata_path

    def _run_task(self):
        if self.project_config.project__package__unpackaged_metadata_path is None:
            self.logger.info(
                "No unpackaged metadata path configured. Skipping metadata deployment task."
            )
            return

        if self.options.get("path") is None:
            self.logger.warning(
                "Unpackaged metadata path is not set. Skipping metadata deployment task."
            )
            return

        try:
            super(DeployUnpackagedMetadata, self)._run_task()
        except Exception as e:
            raise e
        finally:
            clean_temp_directory(self.options["path"])



import typing as T

from snowfakery.cci_mapping_files.declaration_parser import SObjectRuleDeclaration

from cumulusci.tasks.bulkdata.extract_dataset_utils.calculate_dependencies import (
    SObjDependency,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.dependency_map import DependencyMap
from cumulusci.tasks.bulkdata.generate_mapping_utils.mapping_transforms import (
    merge_matching_steps,
    recategorize_lookups,
    rename_record_type_fields,
    sort_steps,
)
from cumulusci.tasks.bulkdata.mapping_parser import MappingStep
from cumulusci.utils.collections import OrderedSet, OrderedSetType

# from ..extract_dataset_utils.extract_yml import ExtractDeclaration
from .mapping_generator_post_processes import add_after_statements

# Note that the code in this file is also used by Snowfakery, so
# changes need to be coordinated.
#
# It is also intended to someday be used by CCI to generate
# Load Mapping Files from Load Declarations.


# Code adapted from Snowfakery
def generate_load_mapping_file(
    mapping_steps: T.Sequence[MappingStep],
    intertable_dependencies: OrderedSetType[SObjDependency],
    load_declarations: T.Sequence[SObjectRuleDeclaration] = (),
) -> T.Dict[str, dict]:
    """Generate a mapping file in optimal order with forward references etc.

    Input is a set of a tables, dependencies between them and load declarations"""

    load_declarations = list(load_declarations) or []
    declared_dependencies = collect_user_specified_dependencies(load_declarations)
    table_names = OrderedSet(mapping.sf_object for mapping in mapping_steps)
    depmap = DependencyMap(
        table_names, intertable_dependencies.union(declared_dependencies)
    )

    # Merge similar steps (output step count will be equal or less)
    mapping_steps = merge_matching_steps(mapping_steps, depmap)

    # Sort steps according to dependencies
    mapping_steps = sort_steps(mapping_steps, depmap)

    # Normalize various forms of record type fields
    mapping_steps = rename_record_type_fields(mapping_steps, depmap)

    # Detect which fields are lookups and properly set them up
    mapping_steps = recategorize_lookups(mapping_steps, depmap)

    # Apply user-specified declarations
    mapping_steps = apply_declarations(mapping_steps, load_declarations)

    mappings_dict = mappings_as_dicts(mapping_steps, depmap)
    add_after_statements(mappings_dict)
    return mappings_dict

    # note that it is entirely possible for e.g.
    # M=10 input mapping steps to relate to N=3 tables which generate P=7 load steps.
    #
    # This invariiant will always hold
    #
    # N <= P <= M
    #
    # e.g. 10 Snowfakery Templates generating Accounts, Contacts, Opportunities
    # But Accounts have 5 different update keys and the other two have none.


def collect_user_specified_dependencies(
    declarations: T.Iterable[SObjectRuleDeclaration],
) -> OrderedSetType[SObjDependency]:
    """The user can specify load order dependencies."""
    declared_dependencies = [decl for decl in declarations if decl.load_after]
    return OrderedSet(
        [
            SObjDependency(decl.sf_object, decl.load_after, "DECLARED", True)
            for decl in declared_dependencies
        ]
    )


def mappings_as_dicts(
    load_steps: T.List[MappingStep],
    depmap: DependencyMap,
) -> T.Mapping[str, T.Dict]:
    """Generate mapping.yml data structures."""
    mappings = {}
    for mapping_step in load_steps:
        step_name = f"{mapping_step.action.value.title()} {mapping_step.sf_object}"
        # Will be used by Snowfakery to generate Upsert mapping files
        if mapping_step.update_key:
            step_name += f" {'_'.join(mapping_step.update_key)}"

        assert step_name not in mappings

        mappings[step_name] = mapping_step.dict(
            by_alias=True,
            exclude_defaults=True,
        )
    return mappings


def apply_declarations(
    mapping_steps: T.List[MappingStep],
    load_declarations: T.List[SObjectRuleDeclaration] = None,
) -> T.List[MappingStep]:
    """Apply user-specified declarations.
    Will be used when this code is used to generate load mappings
    from schemas and SQL instead of from ExtractDeclarations."""

    load_declarations = {decl.sf_object: decl for decl in load_declarations}

    def doit(mapping_step: MappingStep) -> MappingStep:
        if sobject_declarations := load_declarations.get(mapping_step.sf_object):
            return MappingStep(
                **{**mapping_step.dict(), **sobject_declarations.as_mapping()}
            )
        else:
            return mapping_step

    return [doit(step) for step in mapping_steps]



from collections import defaultdict
from contextlib import contextmanager

import yaml

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils.fileutils import FSResource, open_fs_resource


class SetTDTMHandlerStatus(BaseSalesforceApiTask):
    """Alter the active status of specific NPSP or EDA TDTM trigger handlers.
    Save a state file to allow the handlers to be restored to their previous status.
    """

    task_options = {
        "handlers": {
            "description": "List of Trigger Handlers (by Class, Object, or 'Class:Object') to affect (defaults to all handlers)."
        },
        "namespace": {
            "description": "The namespace of the Trigger Handler object ('eda' or 'npsp'). The task will apply the namespace if needed."
        },
        "active": {
            "description": "True or False to activate or deactivate trigger handlers."
        },
        "restore_file": {
            "description": "Path to the state file to store or restore the current trigger handler state. Set to False to discard trigger state information. By default the state is cached in an org-specific directory for later restore."
        },
        "restore": {
            "description": "If True, restore the state of Trigger Handlers to that stored in the (specified or default) restore file."
        },
    }
    should_restore = False
    should_save = False

    def _init_options(self, kwargs):
        super(SetTDTMHandlerStatus, self)._init_options(kwargs)
        self.options["handlers"] = process_list_arg(self.options.get("handlers", []))
        self.options["active"] = process_bool_arg(self.options.get("active") or False)
        has_restore_file = (
            self.options.get("restore_file") is not False
            and self.options.get("restore_file") != "False"
        )
        should_restore = process_bool_arg(self.options.get("restore", False))
        if should_restore:
            if not has_restore_file:
                raise TaskOptionsError("Restoring requires a restore file name")
            self.should_restore = True
        else:
            self.should_save = has_restore_file

    @contextmanager
    def _default_restore_resource(self):
        with self.org_config.get_orginfo_cache_dir(self.__module__) as cache:
            yield cache / "trigger_status.yml"

    def _run_task(self):
        restore_path = self.options.get("restore_file")
        if restore_path:
            restore_file = open_fs_resource(restore_path)
        else:
            restore_file = self._default_restore_resource()

        with restore_file as f:
            self._do_trigger_handlers(f)

    def _do_trigger_handlers(self, restore_file: FSResource):
        global_describe = self.sf.describe()
        sobject_names = [x["name"] for x in global_describe["sobjects"]]

        if self.should_restore:
            with restore_file.open("r") as f:
                target_status = yaml.safe_load(f)
                self.options["handlers"] = list(target_status.keys())
        else:
            target_status = defaultdict(lambda: self.options["active"])

        namespace = self.options.get("namespace", "") + "__"
        if f"{namespace}Trigger_Handler__c" in sobject_names:
            pass
        elif "Trigger_Handler__c" in sobject_names:
            namespace = ""
        else:
            raise CumulusCIException(
                "Unable to locate the Trigger Handler sObject. "
                "Ensure the namespace option is set correctly."
            )

        proxy_obj = getattr(self.sf, f"{namespace}Trigger_Handler__c")
        trigger_handlers = self.sf.query(
            f"SELECT Id, {namespace}Class__c, {namespace}Object__c, {namespace}Active__c FROM {namespace}Trigger_Handler__c"
        )

        current_status = {}

        for handler in trigger_handlers.get("records", []):
            class_name = handler[f"{namespace}Class__c"]
            object_name = handler[f"{namespace}Object__c"]
            compound_name = f"{object_name}:{class_name}"

            if not self.options["handlers"] or any(
                [
                    class_name in self.options["handlers"],
                    object_name in self.options["handlers"],
                    compound_name in self.options["handlers"],
                ]
            ):
                current_status[compound_name] = handler[f"{namespace}Active__c"]
                proxy_obj.update(
                    handler["Id"],
                    {f"{namespace}Active__c": target_status[compound_name]},
                )

        if self.should_save:
            with restore_file.open("w") as f:
                yaml.safe_dump(current_status, f)
        elif self.should_restore:
            restore_file.unlink()



import logging
from abc import ABC
from functools import lru_cache
from typing import Optional, Type

from pydantic.v1 import AnyUrl, root_validator

import cumulusci.core.dependencies.base as base_dependency
from cumulusci.core.exceptions import DependencyResolutionError, GithubApiNotFoundError
from cumulusci.utils.git import split_repo_url
from cumulusci.vcs.github.adapter import GitHubRepository

logger = logging.getLogger(__name__)

VCS_GITHUB = "github"


@lru_cache(50)
def get_github_repo(project_config, url) -> GitHubRepository:
    from cumulusci.vcs.github.service import VCSService, get_github_service_for_url

    vcs_service: Optional[VCSService] = get_github_service_for_url(project_config, url)

    if vcs_service is None:
        raise DependencyResolutionError(
            f"Could not find a GitHub service for URL: {url}"
        )

    try:
        repo = vcs_service.get_repository(options={"repository_url": url})
        if repo is None:
            raise GithubApiNotFoundError(f"Get GitHub Repository found None. {url}")
        return repo
    except GithubApiNotFoundError as e:
        raise DependencyResolutionError(
            f"Could not find a GitHub repository at {url}: {e}"
        )


def _validate_github_parameters(values):
    if values.get("repo_owner") or values.get("repo_name"):
        logger.warning(
            "The repo_name and repo_owner keys are deprecated. Please use the github key."
        )

    assert (
        values.get("url")
        or values.get("github")
        or (values.get("repo_owner") and values.get("repo_name"))
    ), "Must specify `github` or `repo_owner` and `repo_name`"

    # Populate the `github` property if not already populated.
    if not values.get("github") and values.get("repo_name"):
        values["github"] = values[
            "url"
        ] = f"https://github.com/{values['repo_owner']}/{values['repo_name']}"
        values.pop("repo_owner")
        values.pop("repo_name")

    return values


def _sync_github_and_url(values):
    # If only github is provided, set url to github
    if values.get("github") and not values.get("url"):
        values["url"] = values["github"]
    # If only url is provided, set github to url
    elif values.get("url") and not values.get("github"):
        values["github"] = values["url"]
    return values


class GitHubDependencyPin(base_dependency.VcsDependencyPin):
    """Model representing a request to pin a GitHub dependency to a specific tag"""

    github: str

    @property
    def vcsTagResolver(self):  # -> Type["AbstractTagResolver"]:
        from cumulusci.core.dependencies.github_resolvers import (  # Circular imports
            GitHubTagResolver,
        )

        return GitHubTagResolver

    @root_validator(pre=True)
    def sync_vcs_and_url(cls, values):
        """Defined vcs should be assigned to url"""
        return _sync_github_and_url(values)


GitHubDependencyPin.update_forward_refs()


class BaseGitHubDependency(base_dependency.BaseVcsDynamicDependency, ABC):
    """Base class for dynamic dependencies that reference a GitHub repo."""

    github: Optional[AnyUrl] = None
    vcs: str = VCS_GITHUB
    pin_class = GitHubDependencyPin

    repo_owner: Optional[str] = None  # Deprecated - use full URL
    repo_name: Optional[str] = None  # Deprecated - use full URL

    @root_validator
    def check_deprecated_fields(cls, values):
        if values.get("repo_owner") or values.get("repo_name"):
            logger.warning(
                "The dependency keys `repo_owner` and `repo_name` are deprecated. Use the full repo URL with the `github` key instead."
            )
        return values

    @root_validator
    def validate_github_parameters(cls, values):
        return _validate_github_parameters(values)

    @root_validator(pre=True)
    def sync_vcs_and_url(cls, values):
        """Defined vcs should be assigned to url"""
        return _sync_github_and_url(values)


class GitHubDynamicSubfolderDependency(
    BaseGitHubDependency, base_dependency.VcsDynamicSubfolderDependency
):
    """A dependency expressed by a reference to a subfolder of a GitHub repo, which needs
    to be resolved to a specific ref. This is always an unmanaged dependency."""

    @property
    def unmanagedVcsDependency(self) -> Type["UnmanagedGitHubRefDependency"]:
        """A human-readable description of the dependency."""
        return UnmanagedGitHubRefDependency


class GitHubDynamicDependency(
    BaseGitHubDependency, base_dependency.VcsDynamicDependency
):
    """A dependency expressed by a reference to a GitHub repo, which needs
    to be resolved to a specific ref and/or package version."""

    @property
    def unmanagedVcsDependency(self) -> Type["UnmanagedGitHubRefDependency"]:
        """A human-readable description of the dependency."""
        return UnmanagedGitHubRefDependency

    def get_repo(self, context, url) -> "GitHubRepository":
        return get_github_repo(context, url)


class UnmanagedGitHubRefDependency(base_dependency.UnmanagedVcsDependency):
    """Static dependency on unmanaged metadata in a specific GitHub ref and subfolder."""

    repo_owner: Optional[str] = None
    repo_name: Optional[str] = None

    # or
    github: Optional[AnyUrl] = None

    def get_repo(self, context, url) -> "GitHubRepository":
        return get_github_repo(context, url)

    @property
    def package_name(self) -> str:
        repo_owner, repo_name = split_repo_url((str(self.github)))
        package_name = f"{repo_owner}/{repo_name} {self.subfolder}"
        return package_name

    @root_validator
    def validate(cls, values):
        return _validate_github_parameters(values)

    @root_validator(pre=True)
    def sync_vcs_and_url(cls, values):
        """Defined vcs should be assigned to url"""

        return _sync_github_and_url(values)



from datetime import date
from pathlib import Path

import pytest

from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils.options import (
    READONLYDICT_ERROR_MSG,
    CCIOptions,
    Field,
    ListOfStringsOption,
    MappingOption,
    ReadOnlyOptions,
)

ORG_ID = "00D000000000001"
USERNAME = "sample@example"


class TaskToTestTypes(BaseTask):
    class Options(CCIOptions):
        the_list: ListOfStringsOption = Field(
            default=[],
            description="A list of strings",
        )
        the_mapping: MappingOption = Field(
            default=[],
            description="A list of strings",
        )
        the_date: date = Field(default=None, description="A date")
        the_bool: bool = Field(default=True, description="A bool")
        the_path: Path = Field(default=None, description="A bool")
        req: int = Field(..., description="Mandatory int")

    parsed_options: Options

    def _run_task(self):
        for key, value in vars(self.parsed_options).items():
            if value:
                print(key, repr(getattr(self.parsed_options, key)))


class TaskWithoutOptions(BaseTask):
    pass


class TestTaskOptionsParsing:
    def setup_class(self):
        self.global_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.global_config, config={"noyaml": True}
        )
        self.org_config = OrgConfig({"username": USERNAME, "org_id": ORG_ID}, "test")
        self.task_config = TaskConfig(
            {
                "options": {
                    "the_list": ["a"],
                    "the_mapping": {"b": "c"},
                    "the_date": date(2019, 1, 30),
                    "the_bool": True,
                    "req": 1,
                }
            }
        )

    def test_noop(self):
        task = TaskToTestTypes(self.project_config, self.task_config, self.org_config)
        task._init_options({})
        assert task.options["the_list"] == ["a"]
        assert task.options["the_mapping"] == {"b": "c"}
        assert task.options["the_date"] == date(2019, 1, 30)
        assert task.options["the_bool"] is True
        assert task.parsed_options.the_list == task.options["the_list"]
        assert task.parsed_options.the_mapping == task.options["the_mapping"]
        assert task.parsed_options.the_date == task.options["the_date"]
        assert task.parsed_options.the_bool == task.options["the_bool"]

    def test_split_list(self):
        task_config = TaskConfig({"options": {"the_list": "a,b,c,d,e", "req": 1}})
        task = TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert task.parsed_options.the_list == ["a", "b", "c", "d", "e"]

    def test_split_dict(self):
        task_config = TaskConfig(
            {"options": {"the_mapping": "a:aa,b:bb,c:cc", "req": 1}}
        )
        task = TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert task.parsed_options.the_mapping == {"a": "aa", "b": "bb", "c": "cc"}

    def test_parse_date(self):
        task_config = TaskConfig({"options": {"the_date": "1972-11-03", "req": 1}})
        task = TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert task.parsed_options.the_date == date(1972, 11, 3)

    def test_parse_bool(self):
        task_config = TaskConfig({"options": {"the_bool": "False", "req": 1}})
        task = TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert task.parsed_options.the_bool is False

    def test_parse_path(self):
        task_config = TaskConfig({"options": {"the_path": __file__, "req": 1}})
        task = TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert task.parsed_options.the_path.exists()

    def test_missing_option(self):
        task_config = TaskConfig({"options": {"the_path": __file__}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "req" in str(e.value)
        assert "required" in str(e.value)

    def test_wrong_option_type(self):
        task_config = TaskConfig({"options": {"the_bool": "Negatory", "req": 1}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "the_bool" in str(e.value)
        assert "boolean" in str(e.value)

    def test_extra_options(self):
        task_config = TaskConfig({"options": {"foo": "bar", "req": 1}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "foo" in str(e.value)
        assert "extra options" in str(e.value)

    def test_exception_in_parsing(self):
        task_config = TaskConfig(
            {"options": {"the_mapping": "aaaaa,bbbb:cccc", "req": 1}}
        )
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "the_mapping" in str(e.value)
        assert "name/value" in str(e.value)

    def test_wrong_type(self):
        task_config = TaskConfig({"options": {"the_path": 5, "req": 1}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "the_path" in str(e.value)
        assert "path" in str(e.value)

    def test_null_for_required(self):
        task_config = TaskConfig({"options": {"the_path": None, "req": None}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "req" in str(e.value)
        assert "none" in str(e.value)

    def test_multiple_errors(self):
        task_config = TaskConfig({"options": {"the_bool": "Nope"}})
        with pytest.raises(TaskOptionsError) as e:
            TaskToTestTypes(self.project_config, task_config, self.org_config)
        assert "the_bool" in str(e.value)
        assert "req" in str(e.value)
        assert "Errors" in str(e.value)

    def test_options_read_only(self):
        # Has an Options class
        task1 = TaskToTestTypes(self.project_config, self.task_config, self.org_config)
        assert isinstance(task1.options, ReadOnlyOptions)
        # Does not have an Options class
        task2 = TaskWithoutOptions(
            self.project_config, self.task_config, self.org_config
        )
        assert isinstance(task2.options, dict)

    def test_init_options__options_read_only_error(self):
        expected_error_msg = READONLYDICT_ERROR_MSG
        task = TaskToTestTypes(self.project_config, self.task_config, self.org_config)
        # Add new option
        with pytest.raises(TaskOptionsError, match=expected_error_msg):
            task.options["new_option"] = "something"
        # Modify existing option
        with pytest.raises(TaskOptionsError, match=expected_error_msg):
            task.options["test_option"] = 456
        # Delete existing option
        with pytest.raises(TaskOptionsError, match=expected_error_msg):
            del task.options["test_option"]
        # Pop existing option
        with pytest.raises(TaskOptionsError, match=expected_error_msg):
            task.options.pop("test_option")



from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class GetPermissionSetAssignments(BaseSalesforceApiTask):
    def _run_task(self):
        query = f"SELECT PermissionSet.Name,PermissionSetGroupId FROM PermissionSetAssignment WHERE AssigneeId = '{self.org_config.user_id}'"

        self.return_values = []
        for result in self.sf.query_all(query)["records"]:
            if result["PermissionSet"]["Name"] not in self.return_values:
                self.return_values.append(result["PermissionSet"]["Name"])
            if result["PermissionSetGroupId"] is not None:
                psg_query = f"SELECT PermissionSet.Name from PermissionSetGroupComponent where PermissionSetGroupId = '{result['PermissionSetGroupId']}'"
                for psg_result in self.sf.query_all(psg_query)["records"]:
                    if (
                        psg_result["PermissionSet"]
                        and psg_result["PermissionSet"]["Name"]
                        not in self.return_values
                    ):
                        self.return_values.append(psg_result["PermissionSet"]["Name"])

        permsets_str = "\n".join(self.return_values)
        self.logger.info(f"Found permission sets assigned:\n{permsets_str}")



# from unittest.mock import MagicMock, patch

import pytest

# import requests
import responses

import cumulusci.vcs.github.service as github
from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.tests.utils import MockUtil
from cumulusci.tests.util import create_project_config
from cumulusci.vcs.github.adapter import GitHubRef, GitHubRepository, GitHubTag

# from github3.exceptions import NotFoundError
# from github3.git import Reference, Tag


class TestAdapter(GithubApiTestMixin):
    @classmethod
    def teardown_method(cls):
        # clear cached repo -> installation mapping
        github.INSTALLATIONS.clear()

    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def project_config(self):
        project_config = create_project_config()

        return project_config

    @pytest.fixture
    @responses.activate
    def repo(self, gh_api, project_config, mock_util):
        mock_util.mock_get_repo()
        git_repo = GitHubRepository(gh_api, project_config)
        return git_repo

    @responses.activate
    def test_get_ref_for_tag(self, repo):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        ref = repo.get_ref_for_tag("tag_SHA")
        assert ref.sha == "tag_SHA"
        assert isinstance(ref, GitHubRef)

    @responses.activate
    def test_get_ref_for_tag__404(self, repo):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError):
            repo.get_ref_for_tag("tag_SHA")

    @responses.activate
    def test_get_tag_by_ref(self, repo, gh_api, mock_util):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", "tag_SHA"),
            status=200,
        )
        ref = repo.get_ref_for_tag("tag_SHA")

        tag = repo.get_tag_by_ref(ref, "beta/1.0")
        assert isinstance(tag, GitHubTag)
        assert tag.sha == "tag_SHA"

    @responses.activate
    def test_get_tag_by_ref__404(self, repo):
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", "tag_SHA"),
            status=404,
        )
        ref = repo.get_ref_for_tag("tag_SHA")

        with pytest.raises(GithubApiNotFoundError):
            repo.get_tag_by_ref(ref, "beta/1.0")

    @responses.activate
    def test_create_tag_success(self, repo, project_config):
        self.init_github()
        project_config.keychain.set_service(
            "github",
            "test_alias",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "token": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )
        responses.add(
            responses.POST,
            self.repo_api_url + "/git/tags",
            json=self._get_expected_tag("release/1.0", "SHA", "SHA"),
            status=201,
        )
        responses.add(
            responses.POST, self.repo_api_url + "/git/refs", json={}, status=201
        )

        tag = repo.create_tag(
            tag_name="v1.0.0",
            message="Test tag",
            sha="test_sha",
            obj_type="commit",
            tagger={"name": "custom_user", "email": "custom_user@example.com"},
        )

        assert isinstance(tag, GitHubTag)
        assert tag.sha == "SHA"
